#-------------------------------------------------------------------------------
# Name:        Medial Axis Generator
# Purpose:     To provide an easy-to-use tool for generating the
#              medial axis of a polygon, to use in teaching or research
# Author:      Barry Joel Kronenfeld
# Created:     February 2014
# Copyright:   (c) bjkronenfeld 2014
# Licence:     GNU Public Licence v.3
#              If you find this useful, I would love to hear about it.
# Notes:       At this time, the end product does not quite reach the edges
#              of the polygon. The shortfall is greater for high-angle vertices
#              (e.g. locations where the polygon boundary is slightly
#              convex, but not very). I'm looking for an efficient way to extend
#              the medial axis to the polygon boundary using arcpy, but haven't
#              found one yet
#-------------------------------------------------------------------------------

# import modules
import arcpy
import arcpy.da
import os
arcpy.env.overwriteOutput=True
##arcpy.env.workspace=r'C:\Users\Ximing\Dropbox\BK_Software\ArcGIS Scripts\MedialAxis\Scratch.gdb'
##arcpy.env.XYTolerance = "0.0000000001 meters"
##arcpy.env.XYResolution = "0.0000000001 meters"

def report(msg):
    print msg
    arcpy.AddMessage(msg)

def divideLinearUnit(linearUnit,divisor):
    # input linearUnit should be string, divisor should be number
    a,b=linearUnit.split(" ")
    a=str(float(a)/divisor)
    return " ".join([a,b])

def generateApproxMedialAxis(inPolyLayer, densifyDistance, outLineLayer):
    # this works in decent time relying only on ArcPy functions
    # however, the results do not come all the way out to the polygon boundary
    report("initializing generateApproxMedialAxis")

    # clear in-memory workspace
 #   arcpy.Delete_management(r'In_Memory')

    # report number of polygons
    polycount=arcpy.GetCount_management(inPolyLayer).getOutput(0)
    report("{} polygons in input".format(polycount))

    # make copy of input
    report("making copy of input...")
    polycopyfile=r'in_memory\polycopy'
    report("copying to {}".format(polycopyfile))
    workingPolyLayer = arcpy.CopyFeatures_management(inPolyLayer,polycopyfile)

    # copy object ID
    numpoly = int(arcpy.GetCount_management(inPolyLayer).getOutput(0))
    polycount=int(arcpy.GetCount_management(workingPolyLayer).getOutput(0))
    report("{} polygons after copying".format(polycount))
    oid=[0]*numpoly
    sc = arcpy.da.SearchCursor(inPolyLayer,["OID@"])
    rowid=0
    for row in sc:
        oid[rowid]=row[0]
        rowid+=1
    del sc
    arcpy.AddField_management(workingPolyLayer,"PolyID","LONG")
    uc = arcpy.da.UpdateCursor(workingPolyLayer,["PolyID"])
    rowid = 0
    for row in uc:
        uc.updateRow([oid[rowid]])
        rowid+=1
    del uc

    # report number of polygons
    polycount=arcpy.GetCount_management(workingPolyLayer).getOutput(0)
    report("Found {} polygons".format(polycount))

    # densify vertices
    report("densifying polygons...")
    arcpy.Densify_edit(workingPolyLayer, "DISTANCE",densifyDistance)

    # create points layer
    report("extracting vertices...")
    workingVerticesLayer = arcpy.FeatureVerticesToPoints_management(workingPolyLayer,r'In_Memory\vertices')

    # show vertex count
    vcount=arcpy.GetCount_management(workingVerticesLayer).getOutput(0)
    report("Found {} vertices".format(vcount))


    # create Thiessen Polygons around points
    report("creating Thiessen polygons...")
    workingThiessen = arcpy.CreateThiessenPolygons_analysis(workingVerticesLayer,r'In_Memory\Thiessen')

    # convert to lines
    report("converting to line feature class...")
    TLines = arcpy.SplitLine_management(workingThiessen, r'in_memory\tlines')

   # delete identical lines - there will be two copies of each
    report("deleting duplicates...")
    arcpy.DeleteIdentical_management(TLines,["SHAPE"])

    # convert to feature layer (otherwise SelectLayerByLocation throws a weird error)
    report("converting to feature layer...")
    T2= arcpy.MakeFeatureLayer_management(TLines,r'in_memory\T2')

    # select Thiessen lines that are not part of polygon boundary (for deletion)
    report("selecting internal lines...")
    arcpy.SelectLayerByLocation_management(T2,"CROSSED_BY_THE_OUTLINE_OF",workingPolyLayer)
    arcpy.SelectLayerByAttribute_management(T2, selection_type='SWITCH_SELECTION')
    arcpy.SelectLayerByLocation_management(T2,'COMPLETELY_WITHIN',workingPolyLayer,selection_type='SUBSET_SELECTION')

    # record this as the first medial axis (we still have to consolidate the lines)
    report("copying selection...")
    ma_spaghetti = arcpy.CopyFeatures_management(T2,r'In_Memory\ma_spaghetti')

##    # now we can unsplit the lines!!!
##    report("assembling lines...")
##    ma_unsplit=arcpy.UnsplitLine_management(ma_spaghetti,r'In_Memory\ma_unsplit')

    # use spatial join to append FID of original polygons
    print "matching to original polygons..."
    ma_byPoly= arcpy.SpatialJoin_analysis(ma_spaghetti,workingPolyLayer,r'In_Memory\ma_byPoly',"JOIN_ONE_TO_MANY", match_option="WITHIN")

    # delete Join_Count and TARGET_FID fields
    report("cleaning up fields...")
    arcpy.DeleteField_management(ma_byPoly,"Join_Count")
    arcpy.DeleteField_management(ma_byPoly,"TARGET_FID")
    arcpy.DeleteField_management(ma_byPoly,"JOIN_FID")

    # copy to output
    report("saving results...")
    report(outLineLayer)
    arcpy.CopyFeatures_management(ma_byPoly,outLineLayer)
##    # debug
##    workinglayer = outLineLayer + "_working"
##    arcpy.CopyFeatures_management(workingPolyLayer,workinglayer)
##    # delete features created
##    arcpy.Delete_management(workingPolyLayer)
##    arcpy.Delete_management(workingThiessen)
##    arcpy.Delete_management(workingVerticesLayer)
##    #arcpy.Delete_management(ma_byPoly)
##    arcpy.Delete_management(ma_spaghetti)
##    #arcpy.Delete_management(ma_dissolve)
##    arcpy.Delete_management(TLines)
##    arcpy.Delete_management(T2)
##    arcpy.Delete_management(ma_unsplit)
    print "Finished script."



if __name__ == '__main__':
    runinIDE=0
    runinArcMap=1
    runmode=runinArcMap

    inPolyLayer=None
    densifyDistance=None
    outLineLayer=None

    if runmode==runinIDE:
        # get inputs (script version)
        inPolyLayer=r'Z:\Class Files\GEG5880\Spring2015\Instructor\Projects\Medial Axis and Networks\Congressional_MA\TestDistrict.shp'
        densifyDistance="1000 meters"
        outLineLayer=r'C:\sync\teaching\GIS_Modeling\Python\CoursePackage\SampleData\Strahler.gdb\testma'

    if runmode==runinArcMap:
        # get inputs (tool version)
        inPolyLayer=arcpy.GetParameterAsText(0)
        densifyDistance=arcpy.GetParameterAsText(1)
        outLineLayer=arcpy.GetParameterAsText(2)

    # generate medial axis
    generateApproxMedialAxis(inPolyLayer, densifyDistance,outLineLayer)

