#-------------------------------------------------------------------------------
# Name:        Strahler Stream Order
# Purpose:     Calculate Strahler Stream Orders for a River Network
# Author:      GIS Modeling Class, Spring 2013
#              Barry Kronenfeld, Dian Nurhayati, Clayton Ballerine, Brian Howard, Justin Pinnell, Sanjiv Vajjala
# Created:     Spring 2013
# Copyright:   (c) Barry Kronenfeld 2013
# Licence:     GNU Publiic License
#-------------------------------------------------------------------------------

import arcpy
import arcpy.da
arcpy.env.overwriteOutput=True
finishedCount=0

def main(inputStreamSegs, startSegment, StrahlerField):
    reportMessage("Reading inputs...")
    arcpy.Delete_management("in_memory")

    # copy input and delete fields
    streams=fcWithOutFields(inputStreamSegs)

    # Create point feature class representing network nodes
    # Feature Vertices to Points tool
    print "Running Feature Vertices to Points tool..."
    inputJunctions=arcpy.FeatureVerticesToPoints_management(streams, r'in_memory\nodes',"BOTH_ENDS")

    # Delete identical points
    print "Running Delete Identical tool..."
    arcpy.AddXY_management(inputJunctions)
    fields=["POINT_X","POINT_Y"]
    arcpy.DeleteIdentical_management(inputJunctions, fields)

    # get adjacency index
    adjJunction, adjSegment=adjacencyIndex(streams,inputJunctions)

    # create result list to contain strahler stream order values
    maxOID=len(adjJunction)-1
    S=[0]*(maxOID+1)
    # create boolean checked list to look for loops
    checked=[False]*(maxOID+1)

    # get first upstream Node
    upstreamNode=firstUpstreamNode(startSegment,adjJunction,adjSegment)
    print "The upstream node of segment " + str(startSegment) + " is " + str(upstreamNode)

    # calculate Strahler stream order
    global finishedCount
    finishedCount=0
    calcStrahler(adjSegment,adjJunction,startSegment, upstreamNode,S, checked)

    # place results into new field in feature class
    placeResults(S,streams, strahlerField)

    # clear memory
    arcpy.Delete_management("in_memory")

def fcWithOutFields(fc):
    # make a copy of feature class without fields
    # Describe the input (need to test the dataset and data types)
    desc = arcpy.Describe(fc)

    # Make a copy of the input (so we can mantain the original as is)
    copyname=arcpy.ValidateTableName('fccopy','in_memory')
    copyname = 'in_memory/' + copyname
    reportMessage(copyname)
    fccopy= arcpy.CopyFeatures_management(fc, copyname)

    # Use ListFields to get a list of field objects
    fieldObjList = arcpy.ListFields(fccopy)

    # Create an empty list that will be populated with field names
    fieldNameList = []

    # For each field in the object list, add the field name to the
    #  name list.  If the field is required, exclude it, to prevent errors
    for field in fieldObjList:
        if not field.required:
            fieldNameList.append(field.name)

    # dBASE tables require a field other than an OID and Shape.  If this is
    #  the case, retain an extra field (the first one in the original list)
    if desc.dataType in ["ShapeFile", "DbaseTable"]:
        fieldNameList = fieldNameList[1:]

    # Execute DeleteField to delete all fields in the field list.
    arcpy.DeleteField_management(fccopy, fieldNameList)

    # return result
    return fccopy

def adjacencyIndex(segs, junctions):
    reportMessage("creating adjacency index...")


    # Run Spatial Join
    # JOIN_FID is NodeID
    # TARGET_FID is SegmentID
    print "Running Spatial Join..."
    inputJoinFeat= arcpy.SpatialJoin_analysis(segs,junctions,"in_memory\inJoin",'JOIN_ONE_TO_MANY')

    # get number of stream segments so we know how big
    # our result list needs to be
    print "Getting number of stream segments..."
    countCursor=arcpy.da.SearchCursor(segs, "ObjectID")
    segCount=0
    maxOID =0
    for row in countCursor:
        segCount+=1
        if row[0]>maxOID:
            maxOID=row[0]
    reportMessage("Found " + str(segCount) + " segments.")

    print "Getting number of junctions..."
    junctionCountCursor = arcpy.da.SearchCursor(junctions,"ObjectID")
    maxJOID=0
    junctionCount=0
    for row in junctionCountCursor:
        junctionCount+=1
        if row[0]>maxJOID:
            maxJOID=row[0]

    # record junctions associated with each segment
    # in list of lists
    adjJunction=[[]]*(maxOID+1)
    for i in range(maxOID+1):
        adjJunction[i]=[]
    adjSegment=[[]]*(maxJOID+1)
    for i in range(maxJOID+1):
        adjSegment[i]=[]

    adjCursor=arcpy.SearchCursor(inputJoinFeat,
            fields="JOIN_FID; TARGET_FID")
    # JOIN_FID is NodeID
    # TARGET_FID is SegmentID
    for row in adjCursor:
        junction=row.getValue('JOIN_FID')
        segment=row.getValue('TARGET_FID')
        adjJunction[segment].append(junction)
        adjSegment[junction].append(segment)

    return adjJunction,adjSegment

def placeResults(resultList, intoFC, fieldName):
    # put into input feature class
    print "Adding field..."
    arcpy.AddField_management(intoFC, fieldName,"Integer")
    print "Field added!"
    # get cursor
    updateField=fieldName
    upCursor=arcpy.UpdateCursor(intoFC)
    print "Updating values..."
    for row in upCursor:
        featID = row.getValue("ObjectID")
        print featID, resultList[featID]
        row.setValue(updateField, resultList[featID])
        upCursor.updateRow(row)
    print "all values successfully updated!"
    del upCursor
    del row

class loopError(Exception):
    def __init__(self, code):
        self.code=code
    def __str__(self):
        return repr(self.code)


def calcStrahler(adjSegList, adjNodeList, startSeg, upstreamNode, strahlerList, checked):
    # calculates the Strahler stream order recursively
    # the first time this is called, the input should be:
    #  - the furthest downstream segment
    #  - the node just upstream of that segment

    # report progress
    print "working on segment " + str(startSeg)
    global finishedCount
    if finishedCount % 100 ==0:
        reportMessage("Processed " + str(finishedCount) + " stream segments.")
    finishedCount +=1
    # look for loops
    if checked[startSeg]:
        # this indicates a loop error
        try:
            arcpy.AddError("Stream segment with ObjectID " + str(startSeg) + " is part of a cycle.")
            raise loopError(startSeg)
        except loopError as e:
            arcpy.AddError("Error: Stream network cannot have cycles. Please remove cycle(s) and try again.")
            #reportMessage("Error: network has a loop involving ObjectID " + str(startSeg))
            #print "network has loop at ObjectID: ", e.code
            sys.exit(0)
    # mark segment as checked
    checked[startSeg]=True
    # get upstream segments
    upSegList=upSeg(startSeg,upstreamNode,adjNodeList,adjSegList)
    # if there are no upstream segments, assign Strahler value of 1
    if len(upSegList)==0:
        strahlerList[startSeg]=1
    # otherwise, assign maximum Strahler value of upstream segments
    else:
        startSegOrder=0
        for seg in upSegList:
            calcStrahler(adjSegList,adjNodeList,seg,upNode(upstreamNode,seg,adjNodeList,adjSegList),strahlerList, checked)
            if strahlerList[seg]==startSegOrder:
                startSegOrder+=1
            if strahlerList[seg]>startSegOrder:
                startSegOrder=strahlerList[seg]
        strahlerList[startSeg]=startSegOrder
    # show progress
    print "Segment " + str(startSeg) + " is order " + str(strahlerList[startSeg])


def firstUpstreamNode(downstreamSegment, adjNodeList, adjSegList):
    # returns the node immediately upstream of the input segment
    # note: the input segment MUST be the furthest downstream segment

    # get all nodes adjacent to input segment
    adjNodes=adjNodeList[downstreamSegment]
    # loop through nodes
    for node in adjNodes:
        # get all segments adjacent to current node
        adjSegs = adjSegList[node]
        # count number of rows in table (=number of segments adjacent to node)
        adjSegCount=len(adjSegs)
        # this is the correct node if the number of adjacent segments is more than 1
        if adjSegCount>1:
            return node


def upNode(firstNode, ofSegment, adjJunctionList,adjSegmentList):
    # Given one node adjacent to a segment, returns the other node
    curAdjNodeList=adjJunctionList[ofSegment]
    adjNodeListClone=curAdjNodeList[:]
    adjNodeListClone.remove(firstNode)
    if len(adjNodeListClone)>0:
        return adjNodeListClone[0]
    else:
        return -1

def upSeg(startSeg, upstreamNode, adjJunctionList,adjSegmentList):
    # returns list of segments just upstream of input segment
    curAdjSegList = adjSegmentList[upstreamNode]
    adjSegListClone = curAdjSegList[:]
    adjSegListClone.remove(startSeg)
    return adjSegListClone
    # show progress
    print "Upstream node of segment " + str(startSeg) + " is " + str(upstreamNode)
    print "Upstream segment of segment " + str(startSeg) + " is " + str(R)
    return R

def reportMessage(msg):
    arcpy.AddMessage(msg)
    print msg

if __name__ == '__main__':
        # input parameters
    print "Getting input parameters..."
    runinIDE=0
    runinArcMap=1
    runenv=runinIDE

    inputStreamSegs=None
    startSegment=None
    strahlerField=None

    if runenv==runinArcMap:
        inputStreamSegs=arcpy.GetParameterAsText(0)
        startSegment=arcpy.GetParameter(1)
        strahlerField=arcpy.GetParameterAsText(2)

    if runenv==runinIDE:
        inputStreamSegs = r'C:\sync\teaching\GIS_Modeling\Python\CoursePackage\SampleData\Strahler.gdb\Final2'
        startSegment=27673
        strahlerField="S3"

    main(inputStreamSegs,startSegment,strahlerField)
