#-------------------------------------------------------------------------------
# Name:        Strahler Stream Order Calculator
# Purpose:     Calculate Strahler Stream Order on Vector Line Data
# Assumptions: Line features are topologically correct, and meet only at endpoints.
#
# Author:      Barry J Kronenfeld
#
# Created:     11/04/2015
# Copyright:   (c) Barry J Kronenfeld 2015
# Licence:     <your licence>
#-------------------------------------------------------------------------------

import arcpy
arcpy.env.ov=True
import LineAdjacencyTable

def reportMessage(msg):
    print msg
    arcpy.AddMessage(msg)

def sideNbLists(id,adj):
    # returns a list of up to two lists
    # each containing the neighbors on one side
    # does not return list if it is empty
    side1nbs=[]
    side2nbs=[]
    for nb in adj[id]:
        if len(side1nbs)==0:
            side1nbs.append(nb)
        else:
            if nb in adj[side1nbs[0]]:
                side1nbs.append(nb)
            else:
                side2nbs.append(nb)
    # create result list, checking that side lists are not
    # empty before adding them to results
    r=[]
    if len(side1nbs)>0:
        r.append(side1nbs)
    if len(side2nbs)>0:
        r.append(side2nbs)
    return r

def getFirstWorklist(adj,selids):
    # returns list of segments that are not selected
    # for which all neighbors are also neighbors of each other
    r=[]
    # get list of line ids
    idlist=range(len(adj))
    # loop through stream segments
    for id in idlist:
        # check that id is not in selids
        if not id in selids:
            # assume valid unless proved otherwise
            valid=True
            # get neighbor list
            nblist=adj[id]
            # loop through pairs of neighbors
            for nb1 in nblist:
                for nb2 in nblist:
                    if nb1<>nb2:
                        nb1nb=adj[nb1]
                        if not nb2 in nb1nb:
                            valid=False
            # add to list if valid
            if valid:
                r.append(id)
    return r

def allOrdered(idlist,order):
    # returns true if all ids in idlist are ordered
    # false otherwise
    result=True
    for id in idlist:
        if order[id]==-1:
            result=False
    return result



def getNextWorklist(worklist,order,adj):
    # returns list of segments to place in worklist
    # a segment is placed in the worklist if
    # all of the neighbors on one side are ordered
    # for efficiency, candidate neighbors are
    # determined from neighbors of previous worklist

    r=[]
    # get list of candidate line ids
    candidates=[]
    for previd in worklist:
        for prevadj in adj[previd]:
            if order[prevadj]==-1:
                candidates.append(prevadj)
    candidates=list(set(candidates))
    # loop through stream segments
    for id in candidates:
        # check if order hasn't been assigned
        if order[id]==-1:
            # get neighbor lists
            nblists=sideNbLists(id,adj)
            # assume there is no side with all ordered
            # neighbors until proven otherwise
            addtoworklist=False
            # work both sides
            for sidenblist in nblists:
                if allOrdered(sidenblist,order)==True:
                    addtoworklist=True
            # add to work list if a side with all
            # ordered neighbors has been found
            if addtoworklist==True:
                r.append(id)
    return r



def getSubsequentWorklist(order,adj):
    # OLD - KEEPING IN CASE getNextWorklist fails
    # returns list of segments with no stream order, and
    # at least 1 neighbor with a stream order, and either
    # (a) only one unordered neighbor, or
    # (b) no unordered neighbors adjacent
    #     to an ordered neighbor

    r=[]
    # get list of line ids
    if len(order)<>len(adj):
        reportMessage("Error: list lengths don't match!")
    idlist=range(len(order))
    # loop through stream segments
    for id in idlist:
        # check if order hasn't been assigned
        if order[id]==-1:
            # get neighbor list
            nblist=adj[id]
            # get number of ordered and unordered neighbors
            ordereds=[]
            unordereds=[]
            for nb in nblist:
                if order[nb]==-1:
                    unordereds.append(nb)
                else:
                    ordereds.append(nb)
            # determine if it goes into worklist
            if len(ordereds) > 0:
                if len(unordereds)<2:
                    r.append(id)
                else:
                    # assume valid unless proven otherwise
                    valid=True
                    for unordered in unordereds:
                        curadj=adj[unordered]
                        for ordered in ordereds:
                            if ordered in curadj:
                                valid=False
                    if valid==True:
                        r.append(id)
    return r

def assignValues(id,  adj, order,neworder, downseg):
    # assigns stream order for given id
    # and downstream segment id for upstream segments
    # assuming that order for all upstream segments have been cac

    # get side nb lists
    nblists=sideNbLists(id,adj)
    # initialize side orders
    sideOrders=[]
    # work sides
    for nblist in nblists:
        # only work this side if all ordered
        if allOrdered(nblist,order):
            # assign downstream segment id to each upstream segment
            for nb in nblist:
                downseg[nb]=id
            # get list of upstream segment orders
            nborders=[]
            for nb in nblist:
                nborders.append(order[nb])
            # determine max order
            assignorder=max(nborders)
            # determine if there are two at this order
            if nborders.count(assignorder)>1:
                assignorder +=1
            # add to sideOrders
            sideOrders.append(assignorder)
    # determine order of stream from minimum (!) of side orders
    if len(sideOrders)==0:
        neworder[id]=1
    else:
        neworder[id]=min(sideOrders)

def CalculateStreamOrders(linefc,orderfield,downfield):
    # get list of selected features
    selids=[]
    desc=arcpy.Describe(linefc)
    idfield=desc.OIDFieldName
    sc=arcpy.da.SearchCursor(linefc,[idfield])
    for row in sc:
        selids.append(row[0])
    del sc
    # unselect
    if desc.dataType=="FeatureLayer":
        arcpy.SelectLayerByAttribute_management(linefc,"CLEAR_SELECTION")
    # if nothing is selected, everything will be in selids
    # so check for this and set selids to empty list if found
    numfeat=int(arcpy.GetCount_management(linefc).getOutput(0))
    if len(selids)==numfeat:
        selids=[]
    reportMessage(str(len(selids)) + " features selected.")
    reportMessage("selected ids: {}".format(str(selids)))
    # get adjacency table
    adj=LineAdjacencyTable.lineAdjacencyTable(linefc)
    # initialize result lists
    maxid=-1
    sc=arcpy.da.SearchCursor(linefc,[idfield])
    for row in sc:
        if row[0] > maxid:
            maxid=row[0]
    del sc
    order=[-1]*(maxid+1)
    downseg=[-1]*(maxid+1)

    # initialize worklist
    worklist=getFirstWorklist(adj,selids)
    # go until worklist is empty
    iter=1
    while len(worklist)>0:
        reportMessage("Iteration " + str(iter))
        # loop through worklist
        neworder=order[:]
        for id in worklist:
            # assign orders and downstream segment ids
            assignValues(id,adj,order,neworder,downseg)
        # update orders
        order=neworder
        # get new worklist
        worklist=getNextWorklist(worklist,order,adj)
#        worklist=getSubsequentWorklist(order,adj)
        iter+=1

##    # see results
##    for id in range(maxid+1):
##        reportMessage("Stream {} is order {}".format(id,order[id]))
##    maxorder=max(order)
##    maxid=order.index(maxorder)
##    reportMessage("max order: stream {} is order {}".format(str(maxid),str(maxorder)))

    # RECORD RESULTS
    # see if output fields exist
    desc=arcpy.Describe(linefc)
    flist=desc.fields
    hasorderfield=False
    hasdownfield=False
    for f in flist:
        fname=f.name
        if fname==orderfield:
            hasorderfield=True
        if fname==downfield:
            hasdownfield=True

    # if not, create them
    if not hasorderfield:
        arcpy.AddField_management(linefc,orderfield,"LONG")
    if not hasdownfield:
        arcpy.AddField_management(linefc,downfield,"LONG")

    # create cursor
    uc = arcpy.da.UpdateCursor(linefc,[idfield,orderfield,downfield])
    for row in uc:
        row[1]=order[row[0]]
        row[2]=downseg[row[0]]
        uc.updateRow(row)
    del uc

if __name__ == '__main__':
    fromIDE=0
    fromArcMap=1

    runenv=fromArcMap

    linefc=None
    streamorderfield=None
    downstreamidfield=None

    if runenv==fromArcMap:
        linefc=arcpy.GetParameterAsText(0)
        streamorderfield=arcpy.GetParameterAsText(1)
        downstreamidfield=arcpy.GetParameterAsText(2)
    if runenv==fromIDE:
        linefc=r'C:\sync\teaching\GIS_Modeling\Python\CoursePackage\SampleData\Strahler.gdb\smallNetwork'
        streamorderfield="Strahler"
        downstreamidfield="downid"

    CalculateStreamOrders(linefc,streamorderfield,downstreamidfield)

    reportMessage("Finished!")