#-------------------------------------------------------------------------------
# Name:        Strahler Streamsheds
# Purpose:     Identifies all streamsheds of a given Strahler order
# Author:      GIS Modeling Class, Spring 2013
#              Barry Kronenfeld, Dian Nurhayati, Clayton Ballerine, Brian Howard, Justin Pinnell, Sanjiv Vajjala
# Created:     Spring 2013
# Copyright:   (c) Barry Kronenfeld 2013
# Licence:     GNU Publiic License
#-------------------------------------------------------------------------------

import arcpy
import arcpy.da
arcpy.env.overwriteOutput=True
arcpy.Delete_management("in_memory")

def reportMessage(msg):
    print msg
    arcpy.AddMessage(msg)

def main():

    # input parameters# input parameters
##    print "Getting input parameters..."
##    inputStreamSegs=r'C:\sync\teaching\GIS_Modeling\Materials\StreamData\NHD\NHDM0101_working.gdb\FlowLines_noLoops'
##    startSegment = 5030
##    order=4
##    strahlerField = "Strahler2"
##    streamshedField = "Shed4"

    inputStreamSegs=arcpy.GetParameterAsText(0)
    startSegment=arcpy.GetParameter(1)
    order=arcpy.GetParameter(2)
    strahlerField=arcpy.GetParameterAsText(3)
    streamshedField=arcpy.GetParameterAsText(4)

    ##startSegment=int(startSegmentUnicode)
    ##order=int(orderUnicode)

    # Create point feature class representing network nodes
    # Feature Vertices to Points tool
    print "Running Feature Vertices to Points tool..."
    inputJunctions=arcpy.FeatureVerticesToPoints_management(inputStreamSegs, r'in_memory\nodes',"BOTH_ENDS")

    # Delete identical points
    print "Running Delete Identical tool..."
    arcpy.AddXY_management(inputJunctions)
    fields=["POINT_X","POINT_Y"]
    arcpy.DeleteIdentical_management(inputJunctions, fields)

    # create adjacency lists
    adjJunction, adjSegment = adjacencyIndex(inputStreamSegs,inputJunctions)
    maxOID = len(adjJunction)-1
    # create lists to contain strahler stream order values
    # and streamshed IDs
    S=[0]*(maxOID+1)
    Shed=[0]*(maxOID+1)

    # get first upstream Node
    ##    upstreamNode=upNode(startNode,startSegment, inputJoinFeat)
    upstreamNode=firstUpstreamNode(startSegment,adjJunction,adjSegment)
    print "The upstream node of segment " + str(startSegment) + " is " + str(upstreamNode)

    # obtain Strahler stream orders
    reportMessage("reading stream orders...")
    sCurse=arcpy.SearchCursor(inputStreamSegs)
    for sRow in sCurse:
        oid=sRow.getValue("ObjectID")
        strahlerOrder=sRow.getValue(strahlerField)
        S[oid]=strahlerOrder

    # work through streams!!
    reportMessage("working upstream...")
    workStrahler(adjJunction,adjSegment,startSegment, upstreamNode,S, order,Shed)
    print S

    # place results into new field in feature class
    placeResults(Shed,inputStreamSegs, streamshedField)

    # clear memory
    arcpy.Delete_management("in_memory")


def placeResults(resultList, intoFC, fieldName):
    # put into input feature class
    print "Adding field..."
    arcpy.AddField_management(intoFC, fieldName,"Integer")
    print "Field added!"
    # get cursor
    updateField=fieldName
    upCursor=arcpy.UpdateCursor(intoFC)
    print "Updating values..."
    for row in upCursor:
        featID = row.getValue("ObjectID")
        print featID, resultList[featID]
        row.setValue(updateField, resultList[featID])
        upCursor.updateRow(row)
    print "all values successfully updated!"
    del upCursor
    del row

def assignStreamshed(startSeg,upstreamNode,adjJunction,adjSegment, shedList, shedID):
    # assigns the input streamID
    # to the input segment and all segments that are upstream from it

    # assign to current segment
    shedList[startSeg]=shedID
    # get upstream segments
    upSegList=upSeg(startSeg,upstreamNode,adjJunction,adjSegment)
    # loop through them
    for seg in upSegList:
        # recursively assign shed ID
        assignStreamshed(seg,upNode(upstreamNode,seg, adjJunction,adjSegment),adjJunction,adjSegment,shedList,shedID)

def workStrahler(adjJunction,adjSegment, startSeg, upstreamNode, strahlerList, order, shedList):
    # calculates the Strahler stream order recursively
    # the first time this is called, the input should be:
    #  - the furthest downstream segment
    #  - the node just upstream of that segment

    # if Strahler stream order is the input order, assign streamsheds
    if strahlerList[startSeg]==order:
        assignStreamshed(startSeg,upstreamNode,adjJunction,adjSegment,shedList, startSeg)
    else:
        # get upstream segments
        upSegList=upSeg(startSeg,upstreamNode,adjJunction,adjSegment)
        # work the upstream segments
        for seg in upSegList:
            nextUpstreamNode=upNode(upstreamNode,seg,adjJunction,adjSegment)
            workStrahler(adjJunction,adjSegment,seg,nextUpstreamNode,strahlerList,order,shedList)
    # show progress
    print "Segment " + str(startSeg) + " is order " + str(strahlerList[startSeg])

def adjacencyIndex(inputStreamSegs, inputJunctions):
    reportMessage("creating adjacency index...")
    # Run Spatial Join
    # JOIN_FID is NodeID
    # TARGET_FID is SegmentID
    print "Running Spatial Join..."
    inputJoinFeat= arcpy.SpatialJoin_analysis(inputStreamSegs,inputJunctions,"in_memory\inJoin",'JOIN_ONE_TO_MANY')

    # get number of stream segments so we know how big
    # our result list needs to be
    print "Getting number of stream segments..."
    countCursor=arcpy.da.SearchCursor(inputStreamSegs, "ObjectID")
    segCount=0
    maxOID =0
    for row in countCursor:
        segCount+=1
        if row[0]>maxOID:
            maxOID=row[0]
    reportMessage("Found " + str(segCount) + " segments.")

    print "Getting number of junctions..."
    junctionCountCursor = arcpy.da.SearchCursor(inputJunctions,"ObjectID")
    maxJOID=0
    junctionCount=0
    for row in junctionCountCursor:
        junctionCount+=1
        if row[0]>maxJOID:
            maxJOID=row[0]

    # record junctions associated with each segment
    # in list of lists
    adjJunction=[[]]*(maxOID+1)
    for i in range(maxOID+1):
        adjJunction[i]=[]
    adjSegment=[[]]*(maxJOID+1)
    for i in range(maxJOID+1):
        adjSegment[i]=[]

    adjCursor=arcpy.SearchCursor(inputJoinFeat,
            fields="JOIN_FID; TARGET_FID")
    # JOIN_FID is NodeID
    # TARGET_FID is SegmentID
    for row in adjCursor:
        junction=row.getValue('JOIN_FID')
        segment=row.getValue('TARGET_FID')
        adjJunction[segment].append(junction)
        adjSegment[junction].append(segment)

    return adjJunction,adjSegment

def firstUpstreamNode(downstreamSegment, adjNodeList, adjSegList):
    # returns the node immediately upstream of the input segment
    # note: the input segment MUST be the furthest downstream segment

    # get all nodes adjacent to input segment
    adjNodes=adjNodeList[downstreamSegment]
    # loop through nodes
    for node in adjNodes:
        # get all segments adjacent to current node
        adjSegs = adjSegList[node]
        # count number of rows in table (=number of segments adjacent to node)
        adjSegCount=len(adjSegs)
        # this is the correct node if the number of adjacent segments is more than 1
        if adjSegCount>1:
            return node


def upNode(firstNode, ofSegment, adjJunctionList,adjSegmentList):
    # Given one node adjacent to a segment, returns the other node
    curAdjNodeList=adjJunctionList[ofSegment]
    adjNodeListClone=curAdjNodeList[:]
    adjNodeListClone.remove(firstNode)
    if len(adjNodeListClone)>0:
        return adjNodeListClone[0]
    else:
        return -1

def upSeg(startSeg, upstreamNode, adjJunctionList,adjSegmentList):
    # returns list of segments just upstream of input segment
    curAdjSegList = adjSegmentList[upstreamNode]
    adjSegListClone = curAdjSegList[:]
    adjSegListClone.remove(startSeg)
    return adjSegListClone
    # show progress
    print "Upstream node of segment " + str(startSeg) + " is " + str(upstreamNode)
    print "Upstream segment of segment " + str(startSeg) + " is " + str(R)
    return R


##def firstUpstreamNode(downstreamSegment,joinFC):
##    # returns the node immediately upstream of the input segment
##    # note: the input segment MUST be the furthest downstream segment
##
##    # get table containing all nodes adjacent to input segment
##    adjNodeCursor=arcpy.SearchCursor(joinFC,
##            fields="JOIN_FID; TARGET_FID",
##            where_clause="TARGET_FID=" + str(downstreamSegment))
##    # loop through rows in table
##    for nodeRow in adjNodeCursor:
##        # get node ID
##        nodeID=nodeRow.getValue('JOIN_FID')
##        # get table containing all segments adjacent to current node
##        adjSegCursor=arcpy.SearchCursor(joinFC,
##                fields="JOIN_FID; TARGET_FID",
##                where_clause="JOIN_FID=" + str(nodeID))
##        # count number of rows in table (=number of segments adjacent to node)
##        adjSegCount=0
##        for segRow in adjSegCursor:
##            adjSegCount+=1
##        # this is the correct node if the number of adjacent segments is more than 1
##        if adjSegCount>1:
##            return nodeID
##
##
##def upNode(firstNode, ofSegment, joinFC):
##    # Given one node adjacent to a segment, returns the other node
##    adjNodeCursor=arcpy.SearchCursor(joinFC,
##            fields="JOIN_FID; TARGET_FID",
##            where_clause="TARGET_FID=" + str(ofSegment))
##    for row in adjNodeCursor:
##        node=row.getValue('JOIN_FID')
##        if node<>firstNode:
##            return node
##    del adjNodeCursor
##    del row
##
##def upSeg(startSeg, upstreamNode, joinFC):
##    # returns list of segments just upstream of input segment
##    adjSegCursor=arcpy.SearchCursor(joinFC,
##            fields="JOIN_FID; TARGET_FID",
##            where_clause="JOIN_FID=" + str(upstreamNode))
##    countAdjacent=0
##    R=[]
##    for row in adjSegCursor:
##        otherSeg=row.getValue('TARGET_FID')
##        if otherSeg<>startSeg:
##            R.append(otherSeg)
##    del adjSegCursor
##    del row
##    # show progress
##    print "Upstream segment of segment " + str(startSeg) + " is " + str(R)
##    return R

if __name__ == '__main__':
    main()
